#!/usr/bin/python
import argparse
import threading
from threading import Thread
import time
import os  #required for iptables
import platform  #required for os detection
import logging
logging.getLogger("scapy.runtime").setLevel(logging.ERROR)	#supress Scapy warnings`
from scapy.all import *
import sys
sys.path.append('../lib')
import ipaddr
from scanners import *
from fileio import *
from checkip import *

class SnifferThread(Thread):
    def __init__ (self,filter):
        Thread.__init__(self)
        self.filter = filter
    	self._stop = threading.Event()
    def run(self):
	print self.filter
        sniff(filter=self.filter, prn=self.handler, store=0)
    def handler(self,packets):
	if packets.haslayer(IPv6):
		if packets.haslayer(ICMPv6EchoReply):
			res = packets.sprintf("%IPv6.src% %ICMPv6EchoReply.type%")
			results.append(res)
		elif packets.haslayer(ICMPv6DestUnreach):
			res = packets.sprintf("%IPv6.src% %ICMPv6DestUnreach.type% %ICMPv6DestUnreach.code%")
			results.append(res)
		elif packets.haslayer(ICMPv6ParamProblem):
			res = packets.sprintf("%IPv6.src% %ICMPv6ParamProblem.type% %ICMPv6ParamProblem.code%")
			results.append(res)
		elif packets.haslayer(ICMPv6TimeExceeded):
			res = packets.sprintf("%IPv6.src% %ICMPv6TimeExceeded.type% %ICMPv6TimeExceeded.code%")
			results.append(res)
		elif packets.haslayer(TCP):
			res = packets.sprintf("%IPv6.src% %TCP.flags%")
			results.append(res)
		elif packets.haslayer(UDP):
			res = "UDP packet received"
			results.append(res)

def handler1(packets):
	try:
			if packets.nh == 58:
				if packets.payload.type == 136:
					print packets.sprintf("%src% %IPv6.src% %ICMPv6ND_NA.type% %ICMPv6ND_NA.tgt% ")
				elif packets.payload.type == 135:
					print packets.sprintf("%src% %IPv6.src% %ICMPv6ND_NS.type% %ICMPv6ND_NS.tgt% ")
				elif packets.payload.type == 134:
					print packets.sprintf("%src% %IPv6.src% %ICMPv6ND_RA.type%")
				elif packets.payload.type == 133:
					print packets.sprintf("%src% %IPv6.src% %ICMPv6ND_RS.type%")
				elif packets.payload.type == 131:	#Multicast Listener Report
					print packets.sprintf("%src% %IPv6.src% %ICMPv6MLReport.type% %ICMPv6MLReport.mladdr%")
			else:
				print packets.sprintf("%src% %IPv6.src% %IPv6.nh%")
			
	except:
		return  #to avoid errors when a non ipv6 packet is sniffed

def main():
	parser = argparse.ArgumentParser(version='0.2',description='An IPv6 port/network scanner. It can be combined with other modules of the framework.')
	parser.add_argument('interface',  action="store", help="the network interface to use.")
	parser.add_argument('-d', '--destination', action="store", dest="destination", help="the IPv6 address(es) of the target(s) - comma separated.")
	parser.add_argument('-if', '--input_file', action="store", dest="input_file", help="the filename that includes the IPv6 address(es) of the target(s) - one per line")
	parser.add_argument('-of', '--output_file', action="store", dest="output_file", help="the filename where the results will be stored.")
	parser.add_argument('-s', '--source',  action="store", dest="source", default=False, help="the IPv6 address of the sender.")
	parser.add_argument('-gw','--gateway', action="store", dest="gateway", help="a gateway to use (only if required).")
	parser.add_argument('-nsol','--neighbor_solicitation', action="store_true", dest="nsol", default=False, help="Display neighbor solicitation results (IPv6 vs MAC addresses). Default: False")
	parser.add_argument('-pn','--ping_scan', action="store_true", dest="pn", default=False, help="perform a ping6 scan.")
	parser.add_argument('-mpn','--multicast_ping_scan', action="store_true", dest="mpn", default=False, help="perform a multicast link-local ping6 scan.")
	parser.add_argument('-sS','--syn_scan', action="store_true", dest="sS", default=False, help="perform a SYN TCP scan (default)")
	parser.add_argument('-sA','--ack_scan', action="store_true", dest="sA", default=False, help="perform an ACK TCP scan")
	parser.add_argument('-sX','--xmas_scan', action="store_true", dest="sX", default=False, help="perform an XMAS TCP scan")
	parser.add_argument('-sR','--reset_scan', action="store_true", dest="sR", default=False, help="perform a RESET TCP scan")
	parser.add_argument('-sF','--fin_scan', action="store_true", dest="sF", default=False, help="perform a FIN TCP scan")
	parser.add_argument('-sN','--null_scan', action="store_true", dest="sN", default=False, help="perform a NULL TCP scan")
	parser.add_argument('-sU','--udp_scan', action="store_true", dest="sU", default=False, help="perform a UDP scan")
	parser.add_argument('-tr','--traceroute', action="store_true", dest="tr", default=False, help="perform ICMPv6 traceroute")
	parser.add_argument('-tr_tcp','--tcp_traceroute', action="store_true", dest="tr_tcp", default=False, help="perform TCP traceroute")
	parser.add_argument('-mttl','--max_ttl', action="store", dest="mttl", default=15, help="maximum ttl for traceroute")
	parser.add_argument('-fD','--frag-data', action="store_true", dest="fD", default=False, help="add an IPv6 Destination Options Extension Header and fragment the IPv6 datagram into n fragments. The size of the IPv6 Destination Options Header is splitted equally in the n-1 fragments. The last fragment is ALWAYS ONLY the layer 4 protocol")
	parser.add_argument('-fD2','--frag-data-2', action="store_true", dest="fD2", default=False, help="add an IPv6 Destination Options Extension Header and fragment the IPv6 datagram into n fragments by miscalculating checksum. The size of the IPv6 Destination Options Header is splitted equally in the n-1 fragments. The last fragment is ALWAYS ONLY the layer 4 protocol")
	parser.add_argument('-fD3','--frag-data-3', action="store_true", dest="fD3", default=False, help="Yet another attack method. Send atomic fragments with multiple fragment headers")
	parser.add_argument('-fD4','--frag-data-4', action="store_true", dest="fD4", default=False, help="Yet another attack method. Send nested fragments with multiple fragment headers")
	parser.add_argument('-se','--size_of_extheader', action="store", dest="size_of_extheader", default=1, help="the size of the additional Extension Header (in octets of bytes)")
	parser.add_argument('-nf','--no_of_fragments', action="store", dest="number_of_fragments", default=False, help="the number of fragments to send")
	parser.add_argument('-nfh','--no_of_fragment_extension_headers', action="store", dest="number_of_fragment_extension_headers", default=False, help="the number of fragment extension headers to send")
	parser.add_argument('-destport','--destination_port', action="store", dest="destport", default=False, help="destination port of a TCP/UDP scan")
	parser.add_argument('-sM','--smart_scan', action="store_true", dest="smart_scan", default=False, help="perform a smart scan")
	parser.add_argument('-iS', '--input_subnet', action="store", dest="input_subnet", help="the IPv6 prefix to use")
	parser.add_argument('-iC', '--input_combinations', action="store", dest="input_combinations", help="the filename where the combimations to use are stored")
	parser.add_argument('-timeout','--timeout_packets', action="store", dest="send_timeout", default=2, help="The timeout (in seconds) if resending of packets is required.")
	parser.add_argument('-stimeout','--sniffer_timeout', action="store", dest="sniffer_timeout", default=5, help="The timeout (in seconds) when the integrated sniffer (IF used) will exit automatically.")
	parser.add_argument('-no_of_retries','--number_of_retries', action="store", dest="number_of_retries", default=2, help="The number of retries when a response is not received.")
	#parser.add_argument('-threads','--number_of_threads', action="store", dest="no_of_threads", default=10, help="The number of threads to use (for multi-threaded operation).")
	parser.add_argument('-rec','--passive_local_reconnaisance', action="store_true", dest="rec", default=False, help="ATTACK 1: Collects passively Neighbor/Router Solicitation/Advertisement Messages. Creates a list of IPv6 addresses / MAC addresses and various other IPv6 related information")
	values = parser.parse_args()
	
	if os.geteuid() != 0:
	      	print "You must be root to run this script."
	      	exit(1)  
	if (not values.rec) and (not values.pn) and (not values.mpn) and (not values.nsol) and (not values.sS) and (not values.sX) and (not values.sA) and (not values.sN and (not values.sR) and (not values.sF) and (not values.sU) and (not values.tr) and (not values.tr_tcp)):
		print "Please, let me know what you want me to do"
		exit(0)  
	
	conf.verb = 0

	if values.source:
		source_ip = values.source
               	if not is_valid_ipv6(source_ip):
			print "Source ip",source_ip, "is not a valid IPv6 address"
			print "Please, fix the errors and come back"
	elif values.interface:
		source_ip = get_my_ip(values.interface)
	else:
		source_ip=addr
	if values.rec:
		print "Start passive sniffing"
		sniff(store=0,filter="ip6", prn=handler1, iface=values.interface)
		exit(0)
	if values.fD or values.fD2 or values.fD3 or values.fD4:
		myfilter = "ip6 dst "+source_ip
		sniffer = SnifferThread(myfilter)
		print "Start sniffing"
		sniffer.daemon=True
		sniffer.start()
	if values.mpn:
		multi_ping_scanner(source_ip,values.interface, values.sniffer_timeout)
		print "\nLink-local IPv6 addresses (MAC/IPv6)"
		print "------------------------------------"
		for r in link_local_addresses:
			print r
		print "\nGlobal IPv6 addresses (MAC/IPv6)"
		print "------------------------------------"
		for r in global_addresses:
			print r
		exit(0)	

	if not (values.destination or values.input_file or values.smart_scan):
		print "You must define your destination(s)"
		exit(0)

	destports=""
	if values.sS or values.sU or values.sA or values.sR or values.sF or values.sN or values.sX:
		if not values.destport:
			print "In port scanning (TCP/UDP) you must define the destination ports. Please, try again."
			exit(0)
		else:
			portlist = []
			if values.destport.find('-')!=-1: # found
				portragne = values.destport.split('-')
	        		for r in xrange(int(portragne[0]), int(portragne[1]) + 1):
						portlist.append(str(r))
				destports = ','.join(portlist)
			else:
				destports=values.destport
	gw_mac=""
	#print some info
	print "The IPv6 address of your sender is:", source_ip
	print "The interface to use is", values.interface
	
	if values.input_file:
		if not os.path.isfile(values.input_file):
 			print '[-] ' + values.input_file + 'does not exist'
 			exit(0)
 		elif not os.access(values.input_file, os.R_OK):
 			print '[-] ' + values.input_file + 'access is denied'
 			exit(0)
		else:
			ip_list=read_ipv6_addresses(values.input_file)
	elif values.smart_scan:
		if not values.input_subnet:
			print "You must provide the IPv6 prefix (/64) to use for the smart scan"
			exit(0)
		elif not values.input_combinations:
			print "You must provide the input filename where the combimations to use for the smart scan are stored"
			exit(0)
		else:
			if not os.path.isfile(values.input_combinations):
 				print '[-] ' + values.input_file + 'does not exist'
 				exit(0)
 			elif not os.access(values.input_combinations, os.R_OK):
 				print '[-] ' + values.input_file + 'access is denied'
 				exit(0)
			else:
				print "The IPv6 subnet ",values.input_subnet,"/64 will be scanned using smart combinations"
				print "Now let's read the file with the IPv6 addresses of the targets"
				ip_list=[]
				for l in read_ipv6_addresses(values.input_combinations):
					ip_list.append(values.input_subnet+l)
				print "Finished reading"
	elif values.destination.find('/')!=-1: # found
		net6 = ipaddr.IPv6Network(values.destination)
		print "Number of Hosts/IPs to scan = ", net6.numhosts-2
		ip_list=list(net6.iterhosts())
	else:
		ip_list = values.destination.split(",")

	valid_addresses=True
	print "Check if the supplied addresses are valid ones"
	for ip in ip_list:
		if not is_valid_ipv6(ip):
			print ip, "is not a valid IPv6 address"
			valid_addresses=False
	print "Checking validity of provided IPv6 addresses has finished"
	if not valid_addresses:
		print "Please, fix the errors and come back"
		exit(0)	
	#else:	
		#print "The IPv6 address(es) of your target(s) is/are:", ip_list

	if values.gateway:
		#targets=ip_list
		gw_mac = configure_routing(source_ip, values.gateway, values.interface)
		print "The MAC address of your gateway is", gw_mac

	print "Let's start scanning"
	print "Press Ctrl-C to terminate before finishing"
	if not values.tr:	
		label="Target","Protocol","Port/Code","Status","Reason"
		results.append(label)
	if values.gateway:
		ether_dst=gw_mac
	if values.tr_tcp:
		if values.gateway:
			ether_dst=gw_mac
		else:
			print "You must define the gateway to traceroute to your target(s)"
			exit(0)
		tcp_traceroute_scanner(source_ip,ip_list,ether_dst,values.interface,values.mttl)
		exit(0)
	for dest in ip_list:
		try:
			#targets=[]
			if not values.gateway:
				ether_dst=find_single_mac(source_ip, dest, values.interface)
			if not ether_dst:
				print dest, "not found"
			else:
				#print "Target:", ether_dst, dest
				if values.nsol:
					print "IPv6 address ",dest, " has MAC adddress ", ether_dst
				if values.fD:
					if values.pn:
						send_fragmented_datagrams(source_ip, dest, ether_dst, icmpv6(""), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sS:
						send_fragmented_datagrams(source_ip, dest,ether_dst, tcp_packet(values.destport, "S"), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sX:
						send_fragmented_datagrams(source_ip, dest,ether_dst, tcp_packet(values.destport, "FPU"), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sR:
						send_fragmented_datagrams(source_ip, dest,ether_dst, tcp_packet(values.destport, "R"), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sF:
						send_fragmented_datagrams(source_ip, dest,ether_dst, tcp_packet(values.destport, "F"), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sA:
						send_fragmented_datagrams(source_ip, dest,ether_dst, tcp_packet(values.destport, "A"), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sN:
						send_fragmented_datagrams(source_ip, dest,ether_dst, tcp_packet(values.destport, ""), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sU:
						send_fragmented_datagrams(source_ip, dest, ether_dst, udp_connect(values.destport), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
				elif values.fD2:
					if values.pn:
						send_fragmented_datagrams_2(source_ip, dest, ether_dst, icmpv6(""), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sS:
						send_fragmented_datagrams_2(source_ip, dest,ether_dst, tcp_packet(values.destport, "S"), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sX:
						send_fragmented_datagrams_2(source_ip, dest,ether_dst, tcp_packet(values.destport, "FPU"), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sR:
						send_fragmented_datagrams_2(source_ip, dest,ether_dst, tcp_packet(values.destport, "R"), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sF:
						send_fragmented_datagrams_2(source_ip, dest,ether_dst, tcp_packet(values.destport, "F"), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sA:
						send_fragmented_datagrams_2(source_ip, dest,ether_dst, tcp_packet(values.destport, "A"), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sN:
						send_fragmented_datagrams_2(source_ip, dest,ether_dst, tcp_packet(values.destport, ""), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
					elif values.sU:
						send_fragmented_datagrams_2(source_ip, dest, ether_dst, udp_connect(values.destport), int(values.size_of_extheader), int(values.number_of_fragments), values.interface)	
				elif values.fD3:
					if values.pn:
						send_fragmented_datagrams_3(source_ip, dest, ether_dst, icmpv6(""), int(values.number_of_fragment_extension_headers), values.interface)	
					elif values.sS:
						send_fragmented_datagrams_3(source_ip, dest,ether_dst, tcp_packet(values.destport, "S"), int(values.number_of_fragment_extension_headers), values.interface)	
					elif values.sX:
						send_fragmented_datagrams_3(source_ip, dest,ether_dst, tcp_packet(values.destport, "FPU"), int(values.number_of_fragment_extension_headers), values.interface)	
					elif values.sR:
						send_fragmented_datagrams_3(source_ip, dest,ether_dst, tcp_packet(values.destport, "R"), int(values.number_of_fragment_extension_headers), values.interface)	
					elif values.sF:
						send_fragmented_datagrams_3(source_ip, dest,ether_dst, tcp_packet(values.destport, "F"), int(values.number_of_fragment_extension_headers), values.interface)	
					elif values.sA:
						send_fragmented_datagrams_3(source_ip, dest,ether_dst, tcp_packet(values.destport, "A"), int(values.number_of_fragment_extension_headers), values.interface)	
					elif values.sN:
						send_fragmented_datagrams_3(source_ip, dest,ether_dst, tcp_packet(values.destport, ""), int(values.number_of_fragment_extension_headers), values.interface)	
					elif values.sU:
						send_fragmented_datagrams_3(source_ip, dest, ether_dst, udp_connect(values.destport), int(values.number_of_fragment_extension_headers), values.interface)	
				elif values.fD4:
					if values.pn:
						send_fragmented_datagrams_4(source_ip, dest, ether_dst, icmpv6(""),  int(values.number_of_fragment_extension_headers), int(values.number_of_fragments), values.interface)	
					elif values.sS:
						send_fragmented_datagrams_4(source_ip, dest,ether_dst, tcp_packet(values.destport, "S"), int(values.number_of_fragment_extension_headers), int(values.number_of_fragments), values.interface)	
					elif values.sX:
						send_fragmented_datagrams_4(source_ip, dest,ether_dst, tcp_packet(values.destport, "FPU"), int(values.number_of_fragment_extension_headers), int(values.number_of_fragments), values.interface)	
					elif values.sR:
						send_fragmented_datagrams_4(source_ip, dest,ether_dst, tcp_packet(values.destport, "R"), int(values.number_of_fragment_extension_headers),int(values.number_of_fragments), values.interface)	
					elif values.sF:
						send_fragmented_datagrams_4(source_ip, dest,ether_dst, tcp_packet(values.destport, "F"), int(values.number_of_fragment_extension_headers),int(values.number_of_fragments), values.interface)	
					elif values.sA:
						send_fragmented_datagrams_4(source_ip, dest,ether_dst, tcp_packet(values.destport, "A"), int(values.number_of_fragment_extension_headers), int(values.number_of_fragments), values.interface)	
					elif values.sN:
						send_fragmented_datagrams_4(source_ip, dest,ether_dst, tcp_packet(values.destport, ""), int(values.number_of_fragment_extension_headers), int(values.number_of_fragments), values.interface)	
					elif values.sU:
						send_fragmented_datagrams_4(source_ip, dest, ether_dst, udp_connect(values.destport), int(values.number_of_fragment_extension_headers), int(values.number_of_fragments), values.interface)	
				else:
					if values.pn:
						ping_scanner(source_ip,dest,ether_dst,values.interface,float(values.send_timeout),float(values.number_of_retries))
					elif values.sS:
						tcp_syn_scanner(source_ip,dest,destports,ether_dst,values.interface,float(values.send_timeout),float(values.number_of_retries))
					elif values.sX:
						tcp_gen_scanner(source_ip,dest,destports,ether_dst,values.interface,float(values.send_timeout), "FPU",float(values.number_of_retries))
					elif values.sR:
						tcp_gen_scanner(source_ip,dest,destports,ether_dst,values.interface,float(values.send_timeout), "R",float(values.number_of_retries))
					elif values.sF:
						tcp_gen_scanner(source_ip,dest,destports,ether_dst,values.interface,float(values.send_timeout),"F",float(values.number_of_retries))
					elif values.sA:
						tcp_gen_scanner(source_ip,dest,destports,ether_dst,values.interface,float(values.send_timeout), "A",float(values.number_of_retries))
					elif values.sN:
						tcp_gen_scanner(source_ip,dest,destports,ether_dst,values.interface,float(values.send_timeout), "",float(values.number_of_retries))
					elif values.sU:
						udp_scanner(source_ip,dest,destports,ether_dst,values.interface, float(values.send_timeout),float(values.number_of_retries))
					elif values.tr:
						traceroute_scanner(source_ip,dest,ether_dst,values.interface,values.mttl)
					elif values.tr_tcp:
						tcp_traceroute_scanner(source_ip,dest,ether_dst,values.interface,values.mttl)
		except KeyboardInterrupt:
    			print "Exiting on user's request..."
			print "Scanning inComplete!"
			print "Results up to now!"
			print "======================================================"
			for r in results:
				print r
    			exit(1)	

	if values.pn or values.sS or values.sA or values.sF or values.sX or values.sN or values.sR or values.tr or values.sU:	#or values.sT
		print "Scanning Complete!"
		print "======================================================"
		if values.tr:
			i = 0
			for r in results:
				i = i +1
				tr_route = str(i) + " "
				for res in r:
					tr_route = tr_route + " " + res
 				print tr_route
		else:
			for r in results:
				print r
		if values.output_file:
			f = open(values.output_file,'w')
			for r in results:
				f.write(str(r)+"\n")	
			f.close()

if __name__ == '__main__':
    main()
