#!/usr/bin/python
# Filename: create_layers.py

from scapy.all import * 
mymac=""

def get_my_ip(interface):
	myip=""
	try:
		mymac = get_if_hwaddr(interface)  #my MAC address
		for ifaces in in6_getifaddr(): 	#in6_getifaddr()  #return a list of IPs - ifaces, etc
			if ifaces[2]==interface:
				if not myip:
					myip=ifaces[0]
				elif myip[0:6] == "fe80::":
					myip=ifaces[0]	
		return myip
	except:
		print "The interface",interface,"does not exist. Please, try again."
		exit(0)

def find_single_mac(source, destination, interface):
	p=neighsol(destination,source,interface,0)
	if p:
		return p.lladdr

def configure_routing(source, gateway, interface):
	p=neighsol(gateway,source,interface,0)
	conf.route6.add(dst="::/0", gw=gateway, dev=interface)
	try:
		return p.lladdr
	except:
		print "Gateway ",gateway," was not found"
		exit(0)

def icmpv6(payload):
	icmpid=random.randrange(1,65535,1)  #generate a random ICMPv6 id
	header=ICMPv6EchoRequest(data=payload,id=icmpid)
	return header	

def udp_connect(destport):
	source_port=random.randrange(1024,65535,1)
	header = UDP(sport=source_port, dport=[int(x) for x in destport.split(",")])
	return header

def tcp_packet(destport, tcp_flags):
	my_seq_number=random.randrange(0,2*65535,1)
	source_port=random.randrange(1024,65535,1)
	header = TCP(sport=source_port, dport=[int(x) for x in destport.split(",")], seq=my_seq_number, flags=tcp_flags)
	return header

def frag_datagram(layer3_payload, no_of_fragments, next_header, layer3_length, layer4_length):
	datagram=[]
	myid=random.randrange(1,4294967296,1)  #generate a random fragmentation id
	if no_of_fragments>1:
		size_of_fragments = (int(layer3_length))/(8*no_of_fragments) # divided by 8 to get octets of bytes
		for i in range(1, no_of_fragments):
			exheader=IPv6ExtHdrFragment(offset=(i-1)*size_of_fragments,m=1,id=myid, nh=next_header)/layer3_payload[(i-1)*size_of_fragments*8:i*size_of_fragments*8]
			datagram.append(exheader)
		exheader=IPv6ExtHdrFragment(offset=(no_of_fragments-1)*size_of_fragments,m=0,id=myid, nh=next_header)/layer3_payload[(no_of_fragments-1)*size_of_fragments*8:]
		datagram.append(exheader)
		return datagram
	elif no_of_fragments==1:
		exheader=IPv6ExtHdrFragment(offset=0,m=0,id=myid, nh=next_header)/layer3_payload
		datagram.append(exheader)
		return datagram
	
def add_dest_opt_and_fragment(source, destination, layer4, size_of_extheader, number_of_fragments):
	extheader=IPv6ExtHdrDestOpt(options=PadN(optdata='\101'*(size_of_extheader-1)*8))
	datagram=IPv6(src=source,dst=destination)/extheader/layer4
	layer3_payload=str(datagram.payload)
	fragments=frag_datagram(layer3_payload,int(number_of_fragments),60, len(layer3_payload),len(layer4))
	return fragments

def add_fragment_header_and_fragment(source, destination, layer4, number_of_fragment_extension_headers, number_of_fragments):
	extheader=add_fragment_header(number_of_fragment_extension_headers)
	datagram=IPv6(src=source,dst=destination)/extheader/layer4
	layer3_payload=str(datagram.payload)
	fragments=frag_datagram(layer3_payload,int(number_of_fragments),44, len(layer3_payload),len(layer4))
	return fragments

def add_fragment_header(number_of_fragment_headers):
	myid=random.randrange(1,4294967296,1)  #generate a random fragmentation id
	ext_headers = IPv6ExtHdrFragment(offset=0, id=myid, m=0)
	for frag in range(0,number_of_fragment_headers-1):
		#myid=random.randrange(1,4294967296,1)  #generate a random fragmentation id
		ext_headers = IPv6ExtHdrFragment(offset=0, id=myid, m=0)/ext_headers
	return ext_headers

version = '0.2'
# End of create_layers.py
