#!/usr/bin/python
# Filename: scanners.py
import threading
import time
from scapy.all import * 
from create_layers import *
import signal

version = '0.2'

results=[]
link_local_addresses=[]
global_addresses=[]

def signal_handler(signal,frame):
	print "Exiting on user's request"
	sys.exit(0)

def handler(packets):
	try:
		if packets.nh == 58:
			if packets.payload.type == 129:
				res=packets.sprintf("%src% %IPv6.src%")
				if packets[IPv6].src[0:4]=="fe80":
					link_local_addresses.append(res)
				else:
					global_addresses.append(res)
	except:
		return  #to avoid errors when a non ipv6 packet is sniffed

def multi_ping_scanner(source,interface, mytimeout):
	print "Alive systems around..."
	class MyThread ( threading.Thread ):
		def run ( self ):
			time.sleep(1)
			for ifaces in in6_getifaddr(): 	#in6_getifaddr()  #return a list of IPs - ifaces, etc
				if ifaces[2]==interface:
					packet=IPv6(src=ifaces[0],dst="ff02::1")/icmpv6("")
					sendp(Ether(dst="33:33:00:00:00:01")/packet,iface=interface)
	MyThread().start()
	try:
		sniff(store=0,filter="ip6", prn=handler,iface=interface,timeout=int(mytimeout))
	except KeyboardInterrupt:
		signal.signal(signal.SIGINT,signal_handler)	
		
def ping_scanner(source,destination,ether_dst,interface,sendtimeout,no_of_retries):
	packet=IPv6(src=source,dst=destination)/icmpv6("")
	try:
		ans,unans=srp(Ether(dst=ether_dst)/packet,iface=interface, retry=no_of_retries,timeout=sendtimeout)
	except:
		print "An exception has occured. Exiting..."
	for s,r in ans:
		if r.payload.payload.type == 1:	#ICMPv6DestUnreach
			res = s[IPv6].dst,"ICMPv6",s[ICMPv6EchoRequest].type,"UNREACHABLE",r.payload.sprintf("%IPv6.src% %ICMPv6DestUnreach.type% %ICMPv6DestUnreach.code%")
			results.append(res)
		else:
			res = s[IPv6].dst,"ICMPv6",s[ICMPv6EchoRequest].type,"REACHABLE",r.sprintf("%IPv6.src% %IPv6.dst% %IPv6.nh% %ICMPv6EchoReply.type% (id=%ICMPv6EchoReply.id%)")
			results.append(res)

def tcp_syn_scanner(source,destination,mydestports,ether_address,interface,sendtimeout,no_of_retries):
	packet=IPv6(src=source,dst=destination)/tcp_packet(mydestports, "S")
	ans,unans=srp(Ether(dst=ether_address)/packet,iface=interface, retry=no_of_retries,timeout=sendtimeout)
	for s,r in ans:
		if r[IPv6].nh==06: #06 ==TCP
			if r[TCP].flags==18: # 18="SA"
				res = s[IPv6].dst,"TCP",s[TCP].dport,"OPEN","SYN/ACK"
				results.append(res)
			else: 
				res = s[IPv6].dst,"TCP",s[TCP].dport,"CLOSED",r.sprintf("%IPv6.src% %IPv6.nh%{TCP: port=%TCP.sport%, flags=%TCP.flags%}{ICMPv6DestUnreach: %ICMPv6DestUnreach.type% %ICMPv6DestUnreach.code%}")
				results.append(res)
	for s in unans:
		res = s[IPv6].dst,"TCP",s[TCP].dport,"FILTERED","(no response)"
		results.append(res)

def tcp_gen_scanner(source,destination,mydestports,ether_address,interface,sendtimeout, flags,no_of_retries):
	packet=IPv6(src=source,dst=destination)/tcp_packet(mydestports, flags)
	ans,unans=srp(Ether(dst=ether_address)/packet,iface=interface, retry=no_of_retries,timeout=sendtimeout)
	for s,r in ans:
		if r[IPv6].nh==6:
			if s[TCP].dport == r[TCP].sport:
				if  r[TCP].flags==20:
					res = s[IPv6].dst,"TCP",s[TCP].dport,"UNFILTERED","RST/ACK"
					results.append(res)
				else:
					res = s[IPv6].dst,"TCP",s[TCP].dport,"UNFILTERED",r[TCP].flags
					results.append(res)

		else: 
			res = s[IPv6].dst,"TCP",s[TCP].dport,"CLOSED",r.sprintf("%IPv6.src% %IPv6.nh%{TCP: port=%TCP.sport%, flags=%TCP.flags%}{ICMPv6DestUnreach: %ICMPv6DestUnreach.type% %ICMPv6DestUnreach.code%}")
			results.append(res)
	for s in unans:
		res = s[IPv6].dst,"TCP",s[TCP].dport,"FILTERED","(no response)"
		results.append(res)

def udp_scanner(source,destination,mydestports,ether_address,interface,sendtimeout,no_of_retries):
	packet=IPv6(src=source,dst=destination)/udp_connect(mydestports)
	ans,unans=srp(Ether(dst=ether_address)/packet,iface=interface, retry=no_of_retries,timeout=sendtimeout)
	for s,r in ans:
		res = s[IPv6].dst,"UDP",s[UDP].dport,"CLOSED",r.sprintf("%IPv6.src% %IPv6.nh%{ICMPv6DestUnreach: %ICMPv6DestUnreach.type% %ICMPv6DestUnreach.code%}")
		results.append(res)
	for s in unans:
		res = s[IPv6].dst,"UDP",s[UDP].dport,"OPEN/FILTERED""(no response)"
		results.append(res)

def traceroute_scanner(source,destination,ether_addresses,interface,mymaxttl):
	ans,unans=traceroute6(destination,maxttl=mymaxttl,l4=ICMPv6EchoRequest(data=RandString(7)))
	last = source
	res = () 
	for s,r in ans:
		latest_source=r.sprintf("%IPv6.src%")
		if not (last==latest_source):
			if r.haslayer(ICMPv6TimeExceeded):
				res = res + (r.sprintf("%IPv6.src% (%IPv6.hlim%)"),) 
			elif r.haslayer(ICMPv6EchoReply):
				res = res + (r.sprintf("%IPv6.src% (%IPv6.hlim%)"),) 
			#	print r[IPv6].src,r[IPv6].nh
			last=latest_source
	results.append(res)

def tcp_traceroute_scanner(source,destination,ether_addresses,interface,mymaxttl):
	ans,unans=traceroute6(destination,maxttl=mymaxttl)
	str1 = '___'.join(destination)
	filename="> ./"+str1+".svg"
	filename=filename.replace(':', '-')
	ans.graph(target=filename)
	results.append(ans.display())

def send_fragmented_datagrams(source, destination, layer2_addr, layer4, size_of_extheader, number_of_fragments, interface):
	if ((len(layer4)+size_of_extheader) < number_of_fragments):
		print "the number of the fragments should not be bigger than the size of the payload"
	elif number_of_fragments > 0:
		fragments=add_dest_opt_and_fragment(source, destination, layer4, size_of_extheader, number_of_fragments)
		for frag in fragments:
			sendp(Ether(dst=layer2_addr)/IPv6(src=source,dst=destination,nh=44)/frag,iface=interface)
	else:
		sendp(Ether(dst=layer2_addr)/IPv6(src=source,dst=destination)/IPv6ExtHdrDestOpt(options=PadN(optdata='\101'*(size_of_extheader-1)*8))/layer4,iface=interface)

def send_fragmented_datagrams_2(source, destination, layer2_addr, layer4, size_of_extheader, number_of_fragments, interface):
	myid = random.randrange(1,4294967296,1)  #generate a random fragmentation id 
	if ((len(layer4)+size_of_extheader) < number_of_fragments):
		print "the number of the fragments should not be bigger than the size of the payload"
	elif number_of_fragments > 0:
		for i in range(0,number_of_fragments-1):
			packet=IPv6(src=source, dst=destination)/IPv6ExtHdrFragment(offset=i*(size_of_extheader+1), id=myid, m=1)/IPv6ExtHdrDestOpt(nh=60, options=PadN(optdata='\101'*size_of_extheader*8))
			sendp(Ether(dst=layer2_addr)/packet,iface=interface)
		packet=IPv6(src=source, dst=destination)/IPv6ExtHdrFragment(offset=(number_of_fragments-1)*(size_of_extheader+1), id=myid, m=1)/IPv6ExtHdrDestOpt(nh=58,  options=PadN(optdata='\101'*size_of_extheader*8))
		sendp(Ether(dst=layer2_addr)/packet,iface=interface)
		packet=IPv6(src=source, dst=destination)/IPv6ExtHdrFragment(offset=number_of_fragments*(size_of_extheader+1), id=myid, m=0)/layer4
		sendp(Ether(dst=layer2_addr)/packet,iface=interface)
	else:
		sendp(Ether(dst=layer2_addr)/IPv6(src=source,dst=destination)/IPv6ExtHdrDestOpt(options=PadN(optdata='\101'*(size_of_extheader-1)*8))/layer4,iface=interface)
		
def send_fragmented_datagrams_3(source, destination, layer2_addr, layer4, number_of_fragment_headers, interface):
	if number_of_fragment_headers > 0:
		fragment_headers=add_fragment_header(number_of_fragment_headers)
		sendp(Ether(dst=layer2_addr)/IPv6(src=source,dst=destination,nh=44)/fragment_headers/layer4,iface=interface)
	else:
		sendp(Ether(dst=layer2_addr)/IPv6(src=source,dst=destination)/layer4,iface=interface)

def send_fragmented_datagrams_4(source, destination, layer2_addr, layer4, number_of_fragment_headers, number_of_fragments, interface):
	if number_of_fragments > 0:
		myid = random.randrange(1,4294967296,1)  #generate a random fragmentation id 
		if number_of_fragment_headers > 0:
			fragments=add_fragment_header_and_fragment(source, destination, layer4, number_of_fragment_headers, number_of_fragments)
			for frag in fragments:
				sendp(Ether(dst=layer2_addr)/IPv6(src=source,dst=destination,nh=44)/frag,iface=interface)
		else:
			for i in range(0,number_of_fragments-1):
				packet=IPv6(src=source, dst=destination)/IPv6ExtHdrFragment(offset=0, id=myid, m=1)
				sendp(Ether(dst=layer2_addr)/packet,iface=interface)
			packet=IPv6(src=source, dst=destination)/IPv6ExtHdrFragment(offset=0, id=myid, m=0)/layer4
			sendp(Ether(dst=layer2_addr)/packet,iface=interface)
	else:
		if number_of_fragment_headers > 0:
			fragment_headers=add_fragment_header(number_of_fragment_headers)
			sendp(Ether(dst=layer2_addr)/IPv6(src=source,dst=destination,nh=44)/fragment_headers/layer4,iface=interface)
		else:
			sendp(Ether(dst=layer2_addr)/IPv6(src=source,dst=destination)/layer4,iface=interface)
# End of scanners.py
